import xmlsec
from tests import base
from xmlsec import constants as consts

KEY_URL = 'pkcs11;pkcs11:token=test;object=test;pin-value=secret1'


def setUpModule():
    from tests import softhsm_setup

    softhsm_setup.setup()


def tearDownModule():
    from tests import softhsm_setup

    softhsm_setup.teardown()


class TestKeys(base.TestMemoryLeaks):
    def test_del_key(self):
        ctx = xmlsec.SignatureContext(manager=xmlsec.KeysManager())
        ctx.key = xmlsec.Key.from_engine(KEY_URL)
        del ctx.key
        self.assertIsNone(ctx.key)

    def test_set_key(self):
        ctx = xmlsec.SignatureContext(manager=xmlsec.KeysManager())
        ctx.key = xmlsec.Key.from_engine(KEY_URL)
        self.assertIsNotNone(ctx.key)

    def test_sign_bad_args(self):
        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_engine(KEY_URL)
        with self.assertRaises(TypeError):
            ctx.sign('')

    def test_sign_fail(self):
        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_engine(KEY_URL)
        with self.assertRaisesRegex(xmlsec.Error, 'failed to sign'):
            ctx.sign(self.load_xml('sign1-in.xml'))

    def test_sign_case1(self):
        """Should sign a pre-constructed template file using a key from a pkcs11 engine."""
        root = self.load_xml('sign1-in.xml')
        sign = xmlsec.tree.find_node(root, consts.NodeSignature)
        self.assertIsNotNone(sign)

        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_engine(KEY_URL)
        self.assertIsNotNone(ctx.key)
        ctx.key.name = 'rsakey.pem'
        self.assertEqual('rsakey.pem', ctx.key.name)

        ctx.sign(sign)
        self.assertEqual(self.load_xml('sign1-out.xml'), root)



"""Run tests over code examples in the documentation."""

import contextlib
import os
import runpy
from pathlib import Path

import pytest

examples_dir = Path(__file__, '../../doc/source/examples').resolve()
examples = sorted(examples_dir.glob('*.py'))


@contextlib.contextmanager
def cd(where_to):
    """Temporarily change the working directory.

    Restore the current working dir after exiting the context.
    """
    curr = Path.cwd()
    try:
        os.chdir(str(where_to))
        yield
    finally:
        os.chdir(str(curr))


@pytest.mark.parametrize('example', examples, ids=lambda p: p.name)
def test_doc_example(example):
    """Verify example scripts included in the docs are up to date.

    Execute each script in :file:`docs/source/examples`,
    not raising any errors is good enough.
    """
    with cd(example.parent):
        runpy.run_path(str(example))



import multiprocessing
import os
import platform
import subprocess
import sys
import tarfile
import zipfile
from distutils.errors import DistutilsError
from pathlib import Path
from urllib.parse import urljoin
from urllib.request import urlcleanup

from .network import download_lib
from .releases import (
    latest_libiconv_release,
    latest_libxml2_release,
    latest_libxslt_release,
    latest_openssl_release,
    latest_xmlsec_release,
    latest_zlib_release,
)


class CrossCompileInfo:
    def __init__(self, host, arch, compiler):
        self.host = host
        self.arch = arch
        self.compiler = compiler

    @property
    def triplet(self):
        return f'{self.host}-{self.arch}-{self.compiler}'


class StaticBuildHelper:
    def __init__(self, builder):
        self.builder = builder
        self.ext = builder.ext_map['xmlsec']
        self.info = builder.info
        self._prepare_directories()

    def prepare(self, platform_name):
        self.info(f'starting static build on {sys.platform}')
        if platform_name == 'win32':
            self._prepare_windows_build()
        elif 'linux' in platform_name or 'darwin' in platform_name:
            self._prepare_unix_build(platform_name)
        else:
            raise DistutilsError(f'Unsupported static build platform: {platform_name}')

    def _prepare_directories(self):
        buildroot = Path('build', 'tmp')

        prefix_dir = buildroot / 'prefix'
        prefix_dir.mkdir(parents=True, exist_ok=True)
        self.prefix_dir = prefix_dir.absolute()

        build_libs_dir = buildroot / 'libs'
        build_libs_dir.mkdir(exist_ok=True)
        self.build_libs_dir = build_libs_dir

        libs_dir = Path(os.environ.get('PYXMLSEC_LIBS_DIR', 'libs'))
        libs_dir.mkdir(exist_ok=True)
        self.libs_dir = libs_dir
        self.info('{:20} {}'.format('Lib sources in:', self.libs_dir.absolute()))

        self.builder.prefix_dir = self.prefix_dir
        self.builder.build_libs_dir = self.build_libs_dir
        self.builder.libs_dir = self.libs_dir

    def _prepare_windows_build(self):
        release_url = 'https://github.com/mxamin/python-xmlsec-win-binaries/releases/download/2025.07.10/'
        if platform.machine() == 'ARM64':
            suffix = 'win-arm64'
        elif sys.maxsize > 2**32:
            suffix = 'win64'
        else:
            suffix = 'win32'

        libs = [
            f'libxml2-2.11.9-3.{suffix}.zip',
            f'libxslt-1.1.39.{suffix}.zip',
            f'zlib-1.3.1.{suffix}.zip',
            f'iconv-1.18-1.{suffix}.zip',
            f'openssl-3.0.16.pl1.{suffix}.zip',
            f'xmlsec-1.3.7.{suffix}.zip',
        ]

        for libfile in libs:
            url = urljoin(release_url, libfile)
            destfile = self.libs_dir / libfile
            if destfile.is_file():
                self.info(f'Using local copy of "{url}"')
            else:
                self.info(f'Retrieving "{url}" to "{destfile}"')
                urlcleanup()
                download_lib(url, str(destfile))

        for package in self.libs_dir.glob('*.zip'):
            with zipfile.ZipFile(str(package)) as archive:
                destdir = self.build_libs_dir
                archive.extractall(path=str(destdir))

        self.ext.define_macros = [
            ('XMLSEC_CRYPTO', '\\"openssl\\"'),
            ('__XMLSEC_FUNCTION__', '__FUNCTION__'),
            ('XMLSEC_NO_GOST', '1'),
            ('XMLSEC_NO_XKMS', '1'),
            ('XMLSEC_NO_CRYPTO_DYNAMIC_LOADING', '1'),
            ('XMLSEC_CRYPTO_OPENSSL', '1'),
            ('UNICODE', '1'),
            ('_UNICODE', '1'),
            ('LIBXML_ICONV_ENABLED', 1),
            ('LIBXML_STATIC', '1'),
            ('LIBXSLT_STATIC', 1),
            ('XMLSEC_STATIC', 1),
            ('inline', '__inline'),
        ]
        self.ext.libraries = [
            'libxmlsec_a',
            'libxmlsec-openssl_a',
            'libcrypto',
            'iconv_a',
            'libxslt_a',
            'libexslt_a',
            'libxml2_a',
            'zlib',
            'WS2_32',
            'Advapi32',
            'User32',
            'Gdi32',
            'Crypt32',
        ]
        self.ext.library_dirs = [str(path.absolute()) for path in self.build_libs_dir.rglob('lib')]

        includes = [path for path in self.build_libs_dir.rglob('include') if path.is_dir()]
        includes.append(next(path / 'xmlsec' for path in includes if (path / 'xmlsec').is_dir()))
        self.ext.include_dirs = [str(path.absolute()) for path in includes]

    def _prepare_unix_build(self, build_platform):
        self._capture_version_overrides()
        archives = self._ensure_source_archives()
        self._extract_archives(archives)

        env, prefix_arg, ldflags, cross_compile = self._prepare_build_environment(build_platform)
        self._build_dependencies(env, prefix_arg, ldflags, cross_compile)
        self._configure_extension_for_static(build_platform)

    def _capture_version_overrides(self):
        builder = self.builder
        builder.openssl_version = os.environ.get('PYXMLSEC_OPENSSL_VERSION', '3.6.0')
        builder.libiconv_version = os.environ.get('PYXMLSEC_LIBICONV_VERSION', '1.18')
        builder.libxml2_version = os.environ.get('PYXMLSEC_LIBXML2_VERSION', '2.14.6')
        builder.libxslt_version = os.environ.get('PYXMLSEC_LIBXSLT_VERSION', '1.1.43')
        builder.zlib_version = os.environ.get('PYXMLSEC_ZLIB_VERSION', '1.3.1')
        builder.xmlsec1_version = os.environ.get('PYXMLSEC_XMLSEC1_VERSION', '1.3.9')

    def _ensure_source_archives(self):
        return [
            self._ensure_source(
                name='OpenSSL',
                glob='openssl*.tar.gz',
                filename='openssl.tar.gz',
                version=self.builder.openssl_version,
                env_label='PYXMLSEC_OPENSSL_VERSION',
                default_url=latest_openssl_release,
                version_url=lambda v: f'https://api.github.com/repos/openssl/openssl/tarball/openssl-{v}',
            ),
            self._ensure_source(
                name='zlib',
                glob='zlib*.tar.gz',
                filename='zlib.tar.gz',
                version=self.builder.zlib_version,
                env_label='PYXMLSEC_ZLIB_VERSION',
                default_url=latest_zlib_release,
                version_url=lambda v: f'https://zlib.net/fossils/zlib-{v}.tar.gz',
            ),
            self._ensure_source(
                name='libiconv',
                glob='libiconv*.tar.gz',
                filename='libiconv.tar.gz',
                version=self.builder.libiconv_version,
                env_label='PYXMLSEC_LIBICONV_VERSION',
                default_url=latest_libiconv_release,
                version_url=lambda v: f'https://ftpmirror.gnu.org/libiconv/libiconv-{v}.tar.gz',
            ),
            self._ensure_source(
                name='libxml2',
                glob='libxml2*.tar.xz',
                filename='libxml2.tar.xz',
                version=self.builder.libxml2_version,
                env_label='PYXMLSEC_LIBXML2_VERSION',
                default_url=latest_libxml2_release,
                version_url=lambda v: self._libxml_related_url('libxml2', v),
            ),
            self._ensure_source(
                name='libxslt',
                glob='libxslt*.tar.xz',
                filename='libxslt.tar.xz',
                version=self.builder.libxslt_version,
                env_label='PYXMLSEC_LIBXSLT_VERSION',
                default_url=latest_libxslt_release,
                version_url=lambda v: self._libxml_related_url('libxslt', v),
            ),
            self._ensure_source(
                name='xmlsec1',
                glob='xmlsec1*.tar.gz',
                filename='xmlsec1.tar.gz',
                version=self.builder.xmlsec1_version,
                env_label='PYXMLSEC_XMLSEC1_VERSION',
                default_url=latest_xmlsec_release,
                version_url=lambda v: f'https://github.com/lsh123/xmlsec/releases/download/{v}/xmlsec1-{v}.tar.gz',
            ),
        ]

    def _ensure_source(self, name, glob, filename, version, env_label, default_url, version_url):
        archive = next(self.libs_dir.glob(glob), None)
        if archive is not None:
            return archive

        self.info('{:10}: {}'.format(name, 'source tar not found, downloading ...'))
        archive = self.libs_dir / filename
        if version is None:
            url = default_url()
            self.info('{:10}: {}'.format(name, f'{env_label} unset, downloading latest from {url}'))
        else:
            url = version_url(version)
            self.info('{:10}: {}'.format(name, f'{env_label}={version}, downloading from {url}'))
        download_lib(url, str(archive))
        return archive

    def _libxml_related_url(self, lib_name, version):
        version_prefix, _ = version.rsplit('.', 1)
        return f'https://download.gnome.org/sources/{lib_name}/{version_prefix}/{lib_name}-{version}.tar.xz'

    def _extract_archives(self, archives):
        for archive in archives:
            self.info(f'Unpacking {archive.name}')
            try:
                with tarfile.open(str(archive)) as tar:
                    tar.extractall(path=str(self.build_libs_dir))
            except EOFError as error:
                raise DistutilsError(f'Bad {archive.name} downloaded; remove it and try again.') from error

    def _prepare_build_environment(self, build_platform):
        prefix_arg = f'--prefix={self.prefix_dir}'
        env = os.environ.copy()

        cflags = []
        if env.get('CFLAGS'):
            cflags.append(env['CFLAGS'])
        cflags.append('-fPIC')

        ldflags = []
        if env.get('LDFLAGS'):
            ldflags.append(env['LDFLAGS'])

        cross_compile = None
        if build_platform == 'darwin':
            arch = self.builder.plat_name.rsplit('-', 1)[1]
            if arch != platform.machine() and arch in ('x86_64', 'arm64'):
                self.info(f'Cross-compiling for {arch}')
                cflags.append(f'-arch {arch}')
                ldflags.append(f'-arch {arch}')
                cross_compile = CrossCompileInfo('darwin64', arch, 'cc')
            major_version, _ = tuple(map(int, platform.mac_ver()[0].split('.')[:2]))
            if major_version >= 11 and 'MACOSX_DEPLOYMENT_TARGET' not in env:
                env['MACOSX_DEPLOYMENT_TARGET'] = '11.0'

        env['CFLAGS'] = ' '.join(cflags)
        env['LDFLAGS'] = ' '.join(ldflags)
        return env, prefix_arg, ldflags, cross_compile

    def _build_dependencies(self, env, prefix_arg, ldflags, cross_compile):
        self._build_openssl(env, prefix_arg, cross_compile)
        self._build_zlib(env, prefix_arg)

        host_arg = [f'--host={cross_compile.arch}'] if cross_compile else []
        self._build_libiconv(env, prefix_arg, host_arg)
        self._build_libxml2(env, prefix_arg, host_arg)
        self._build_libxslt(env, prefix_arg, host_arg)

        ldflags.append('-lpthread')
        env['LDFLAGS'] = ' '.join(ldflags)
        self._build_xmlsec1(env, prefix_arg, host_arg)

    def _build_openssl(self, env, prefix_arg, cross_compile):
        self.info('Building OpenSSL')
        openssl_dir = next(self.build_libs_dir.glob('openssl-*'))
        openssl_config_cmd = [prefix_arg, 'no-shared', '-fPIC', '--libdir=lib']
        if platform.machine() == 'riscv64':
            # openssl(riscv64): disable ASM to avoid R_RISCV_JAL relocation failure on 3.5.2
            # OpenSSL 3.5.2 enables RISC-V64 AES assembly by default. When we statically
            # link libcrypto alongside xmlsec, the AES asm path triggers a link-time error:
            #   relocation truncated to fit: R_RISCV_JAL against symbol `AES_set_encrypt_key'
            #   in .../libcrypto.a(libcrypto-lib-aes-riscv64.o)
            # This appears to stem from a long-range jump emitted by the AES asm generator
            # (see aes-riscv64.pl around L1069), which can exceed the JAL reach when objects
            # end up far apart in the final static link.
            # As a pragmatic workaround, disable ASM on riscv64 (pass `no-asm`) so the
            # portable C implementation is used. This unblocks the build at the cost of
            # some crypto performance on riscv64 only.
            # Refs:
            # - https://github.com/openssl/openssl/blob/0893a62/crypto/aes/asm/aes-riscv64.pl#L1069
            openssl_config_cmd.append('no-asm')
        if cross_compile:
            openssl_config_cmd.insert(0, './Configure')
            openssl_config_cmd.append(cross_compile.triplet)
        else:
            openssl_config_cmd.insert(0, './config')
        subprocess.check_call(openssl_config_cmd, cwd=str(openssl_dir), env=env)
        subprocess.check_call(['make', f'-j{multiprocessing.cpu_count() + 1}'], cwd=str(openssl_dir), env=env)
        subprocess.check_call(['make', f'-j{multiprocessing.cpu_count() + 1}', 'install_sw'], cwd=str(openssl_dir), env=env)

    def _build_zlib(self, env, prefix_arg):
        self.info('Building zlib')
        zlib_dir = next(self.build_libs_dir.glob('zlib-*'))
        subprocess.check_call(['./configure', prefix_arg], cwd=str(zlib_dir), env=env)
        subprocess.check_call(['make', f'-j{multiprocessing.cpu_count() + 1}'], cwd=str(zlib_dir), env=env)
        subprocess.check_call(['make', f'-j{multiprocessing.cpu_count() + 1}', 'install'], cwd=str(zlib_dir), env=env)

    def _build_libiconv(self, env, prefix_arg, host_arg):
        self.info('Building libiconv')
        libiconv_dir = next(self.build_libs_dir.glob('libiconv-*'))
        subprocess.check_call(
            [
                './configure',
                prefix_arg,
                '--disable-dependency-tracking',
                '--disable-shared',
                *host_arg,
            ],
            cwd=str(libiconv_dir),
            env=env,
        )
        subprocess.check_call(['make', f'-j{multiprocessing.cpu_count() + 1}'], cwd=str(libiconv_dir), env=env)
        subprocess.check_call(['make', f'-j{multiprocessing.cpu_count() + 1}', 'install'], cwd=str(libiconv_dir), env=env)

    def _build_libxml2(self, env, prefix_arg, host_arg):
        self.info('Building LibXML2')
        libxml2_dir = next(self.build_libs_dir.glob('libxml2-*'))
        subprocess.check_call(
            [
                './configure',
                prefix_arg,
                '--disable-dependency-tracking',
                '--disable-shared',
                '--without-lzma',
                '--without-python',
                f'--with-iconv={self.prefix_dir}',
                f'--with-zlib={self.prefix_dir}',
                *host_arg,
            ],
            cwd=str(libxml2_dir),
            env=env,
        )
        subprocess.check_call(['make', f'-j{multiprocessing.cpu_count() + 1}'], cwd=str(libxml2_dir), env=env)
        subprocess.check_call(['make', f'-j{multiprocessing.cpu_count() + 1}', 'install'], cwd=str(libxml2_dir), env=env)

    def _build_libxslt(self, env, prefix_arg, host_arg):
        self.info('Building libxslt')
        libxslt_dir = next(self.build_libs_dir.glob('libxslt-*'))
        subprocess.check_call(
            [
                './configure',
                prefix_arg,
                '--disable-dependency-tracking',
                '--disable-shared',
                '--without-python',
                '--without-crypto',
                f'--with-libxml-prefix={self.prefix_dir}',
                *host_arg,
            ],
            cwd=str(libxslt_dir),
            env=env,
        )
        subprocess.check_call(['make', f'-j{multiprocessing.cpu_count() + 1}'], cwd=str(libxslt_dir), env=env)
        subprocess.check_call(['make', f'-j{multiprocessing.cpu_count() + 1}', 'install'], cwd=str(libxslt_dir), env=env)

    def _build_xmlsec1(self, env, prefix_arg, host_arg):
        self.info('Building xmlsec1')
        xmlsec1_dir = next(self.build_libs_dir.glob('xmlsec1-*'))
        subprocess.check_call(
            [
                './configure',
                prefix_arg,
                '--disable-shared',
                '--disable-gost',
                '--enable-md5',
                '--enable-ripemd160',
                '--disable-crypto-dl',
                '--enable-static=yes',
                '--enable-shared=no',
                '--enable-static-linking=yes',
                '--with-default-crypto=openssl',
                f'--with-openssl={self.prefix_dir}',
                f'--with-libxml={self.prefix_dir}',
                f'--with-libxslt={self.prefix_dir}',
                *host_arg,
            ],
            cwd=str(xmlsec1_dir),
            env=env,
        )
        include_flags = [
            f'-I{self.prefix_dir / "include"}',
            f'-I{self.prefix_dir / "include" / "libxml"}',
        ]
        subprocess.check_call(
            ['make', f'-j{multiprocessing.cpu_count() + 1}', *include_flags],
            cwd=str(xmlsec1_dir),
            env=env,
        )
        subprocess.check_call(['make', f'-j{multiprocessing.cpu_count() + 1}', 'install'], cwd=str(xmlsec1_dir), env=env)

    def _configure_extension_for_static(self, build_platform):
        self.ext.define_macros = [
            ('__XMLSEC_FUNCTION__', '__func__'),
            ('XMLSEC_NO_SIZE_T', None),
            ('XMLSEC_NO_GOST', '1'),
            ('XMLSEC_NO_GOST2012', '1'),
            ('XMLSEC_NO_XKMS', '1'),
            ('XMLSEC_CRYPTO', '\\"openssl\\"'),
            ('XMLSEC_NO_CRYPTO_DYNAMIC_LOADING', '1'),
            ('XMLSEC_CRYPTO_OPENSSL', '1'),
            ('LIBXML_ICONV_ENABLED', 1),
            ('LIBXML_STATIC', 1),
            ('LIBXSLT_STATIC', 1),
            ('XMLSEC_STATIC', 1),
            ('inline', '__inline'),
            ('UNICODE', '1'),
            ('_UNICODE', '1'),
        ]

        self.ext.include_dirs.append(str(self.prefix_dir / 'include'))
        self.ext.include_dirs.extend([str(path.absolute()) for path in (self.prefix_dir / 'include').iterdir() if path.is_dir()])

        self.ext.library_dirs = []
        if build_platform == 'linux':
            self.ext.libraries = ['m', 'rt']
        extra_objects = [
            'libxmlsec1.a',
            'libxslt.a',
            'libxml2.a',
            'libz.a',
            'libxmlsec1-openssl.a',
            'libcrypto.a',
            'libiconv.a',
            'libxmlsec1.a',
        ]
        self.ext.extra_objects = [str(self.prefix_dir / 'lib' / obj) for obj in extra_objects]


__all__ = ('CrossCompileInfo', 'StaticBuildHelper')



def pytest_collection_modifyitems(items):
    """Put the module init test first.

    This way, we implicitly check whether any subsequent test fails because of module reinitialization.
    """

    def module_init_tests_first(item):
        return int('test_xmlsec.py::TestModule::test_reinitialize_module' not in item.nodeid)

    items.sort(key=module_init_tests_first)



"""Test type stubs for correctness where possible."""

import os

import pytest

import xmlsec

black = pytest.importorskip('black')


constants_stub_header = """
import sys
from typing import NamedTuple

if sys.version_info >= (3, 8):
    from typing import Final
else:
    from typing_extensions import Final

class __KeyData(NamedTuple):  # __KeyData type
    href: str
    name: str

class __KeyDataNoHref(NamedTuple):  # __KeyData type
    href: None
    name: str

class __Transform(NamedTuple):  # __Transform type
    href: str
    name: str
    usage: int

class __TransformNoHref(NamedTuple):  # __Transform type
    href: None
    name: str
    usage: int

"""


def gen_constants_stub():
    """Generate contents of the file:`xmlsec/constants.pyi`.

    Simply load all constants at runtime,
    generate appropriate type hint for each constant type.
    """

    def process_constant(name):
        """Generate line in stub file for constant name."""
        obj = getattr(xmlsec.constants, name)
        type_name = type(obj).__name__
        if type_name in ('__KeyData', '__Transform') and obj.href is None:
            type_name += 'NoHref'
        return f'{name}: Final[{type_name}]'

    names = list(sorted(name for name in dir(xmlsec.constants) if not name.startswith('__')))
    lines = [process_constant(name) for name in names]
    return constants_stub_header + os.linesep.join(lines)


def test_xmlsec_constants_stub(request):
    """Generate the stub file for :mod:`xmlsec.constants` from existing code.

    Compare it against the existing stub :file:`xmlsec/constants.pyi`.
    """
    stub = request.config.rootpath / 'src' / 'xmlsec' / 'constants.pyi'
    mode = black.FileMode(target_versions={black.TargetVersion.PY39}, line_length=130, is_pyi=True, string_normalization=False)
    formatted = black.format_file_contents(gen_constants_stub(), fast=False, mode=mode)
    assert formatted == stub.read_text()



import copy
import tempfile

import xmlsec
from tests import base

consts = xmlsec.constants


class TestKeys(base.TestMemoryLeaks):
    def test_key_from_memory(self):
        key = xmlsec.Key.from_memory(self.load('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(key)

    def test_key_from_memory_with_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.Key.from_memory(1, format='')

    def test_key_from_memory_invalid_data(self):
        with self.assertRaisesRegex(xmlsec.Error, '.*cannot load key.*'):
            xmlsec.Key.from_memory(b'foo', format=consts.KeyDataFormatPem)

    def test_key_from_file(self):
        key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(key)

    def test_key_from_file_with_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.Key.from_file(1, format='')

    def test_key_from_invalid_file(self):
        with self.assertRaisesRegex(xmlsec.Error, '.*cannot read key.*'), tempfile.NamedTemporaryFile() as tmpfile:
            tmpfile.write(b'foo')
            xmlsec.Key.from_file(tmpfile.name, format=consts.KeyDataFormatPem)

    def test_key_from_fileobj(self):
        with open(self.path('rsakey.pem'), 'rb') as fobj:
            key = xmlsec.Key.from_file(fobj, format=consts.KeyDataFormatPem)
        self.assertIsNotNone(key)

    def test_key_from_invalid_fileobj(self):
        with tempfile.NamedTemporaryFile(delete=False) as tmpfile:
            tmpfile.write(b'foo')
        with self.assertRaisesRegex(xmlsec.Error, '.*cannot read key.*'), open(tmpfile.name) as fp:
            xmlsec.Key.from_file(fp, format=consts.KeyDataFormatPem)

    def test_generate(self):
        key = xmlsec.Key.generate(klass=consts.KeyDataAes, size=256, type=consts.KeyDataTypeSession)
        self.assertIsNotNone(key)

    def test_generate_with_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.Key.generate(klass='', size='', type='')

    def test_generate_invalid_size(self):
        with self.assertRaisesRegex(xmlsec.Error, '.*cannot generate key.*'):
            xmlsec.Key.generate(klass=consts.KeyDataAes, size=0, type=consts.KeyDataTypeSession)

    def test_from_binary_file(self):
        key = xmlsec.Key.from_binary_file(klass=consts.KeyDataDes, filename=self.path('deskey.bin'))
        self.assertIsNotNone(key)

    def test_from_binary_file_with_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.Key.from_binary_file(klass='', filename=1)

    def test_from_invalid_binary_file(self):
        with self.assertRaisesRegex(xmlsec.Error, '.*cannot read key.*'), tempfile.NamedTemporaryFile() as tmpfile:
            tmpfile.write(b'foo')
            xmlsec.Key.from_binary_file(klass=consts.KeyDataDes, filename=tmpfile.name)

    def test_from_binary_data(self):
        key = xmlsec.Key.from_binary_data(klass=consts.KeyDataDes, data=self.load('deskey.bin'))
        self.assertIsNotNone(key)

    def test_from_binary_data_with_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.Key.from_binary_data(klass='', data=1)

    def test_from_invalid_binary_data(self):
        with self.assertRaisesRegex(xmlsec.Error, '.*cannot read key.*'):
            xmlsec.Key.from_binary_data(klass=consts.KeyDataDes, data=b'')

    def test_load_cert_from_file(self):
        key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(key)
        key.load_cert_from_file(self.path('rsacert.pem'), format=consts.KeyDataFormatPem)

    def test_load_cert_from_file_with_bad_args(self):
        key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(key)
        with self.assertRaises(TypeError):
            key.load_cert_from_file(1, format='')

    def test_load_cert_from_invalid_file(self):
        key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(key)
        with self.assertRaisesRegex(xmlsec.Error, '.*cannot load cert.*'), tempfile.NamedTemporaryFile() as tmpfile:
            tmpfile.write(b'foo')
            key.load_cert_from_file(tmpfile.name, format=consts.KeyDataFormatPem)

    def test_load_cert_from_fileobj(self):
        key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(key)
        with open(self.path('rsacert.pem'), 'rb') as fobj:
            key.load_cert_from_file(fobj, format=consts.KeyDataFormatPem)

    def test_load_cert_from_fileobj_with_bad_args(self):
        key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(key)
        with self.assertRaises(TypeError), open(self.path('rsacert.pem'), 'rb') as fobj:
            key.load_cert_from_file(fobj, format='')

    def test_load_cert_from_invalid_fileobj(self):
        key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(key)
        with tempfile.NamedTemporaryFile(delete=False) as tmpfile:
            tmpfile.write(b'foo')
        with self.assertRaisesRegex(xmlsec.Error, '.*cannot load cert.*'), open(tmpfile.name) as fp:
            key.load_cert_from_file(fp, format=consts.KeyDataFormatPem)

    def test_load_cert_from_memory(self):
        key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(key)
        key.load_cert_from_memory(self.load('rsacert.pem'), format=consts.KeyDataFormatPem)

    def test_load_cert_from_memory_with_bad_args(self):
        key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(key)
        with self.assertRaises(TypeError):
            key.load_cert_from_memory(1, format='')

    def test_load_cert_from_memory_invalid_data(self):
        key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(key)
        with self.assertRaisesRegex(xmlsec.Error, '.*cannot load cert.*'):
            key.load_cert_from_memory(b'', format=consts.KeyDataFormatPem)

    def test_get_name(self):
        key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNone(key.name)

    def test_get_name_invalid_key(self):
        key = xmlsec.Key()
        with self.assertRaisesRegex(ValueError, 'key is not ready'):
            key.name  # noqa: B018

    def test_del_name(self):
        key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        key.name = 'rsakey'
        del key.name
        self.assertIsNone(key.name)

    def test_set_name(self):
        key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        key.name = 'rsakey'
        self.assertEqual('rsakey', key.name)

    def test_set_name_invalid_key(self):
        key = xmlsec.Key()
        with self.assertRaisesRegex(ValueError, 'key is not ready'):
            key.name = 'foo'

    def test_copy(self):
        key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        key2 = copy.copy(key)
        del key
        key2.load_cert_from_file(self.path('rsacert.pem'), format=consts.KeyDataFormatPem)


class TestKeysManager(base.TestMemoryLeaks):
    def test_add_key(self):
        key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        mngr = xmlsec.KeysManager()
        mngr.add_key(key)

    def test_add_key_with_bad_args(self):
        mngr = xmlsec.KeysManager()
        with self.assertRaises(TypeError):
            mngr.add_key('')

    def test_load_cert(self):
        mngr = xmlsec.KeysManager()
        mngr.add_key(xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem))
        mngr.load_cert(self.path('rsacert.pem'), format=consts.KeyDataFormatPem, type=consts.KeyDataTypeTrusted)

    def test_load_cert_with_bad_args(self):
        mngr = xmlsec.KeysManager()
        mngr.add_key(xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem))
        with self.assertRaisesRegex(xmlsec.Error, '.*cannot load cert.*'), tempfile.NamedTemporaryFile() as tmpfile:
            tmpfile.write(b'foo')
            mngr.load_cert(tmpfile.name, format=consts.KeyDataFormatPem, type=consts.KeyDataTypeTrusted)

    def test_load_invalid_cert(self):
        mngr = xmlsec.KeysManager()
        mngr.add_key(xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem))
        with self.assertRaises(TypeError):
            mngr.load_cert(1, format='', type='')

    def test_load_cert_from_memory(self):
        mngr = xmlsec.KeysManager()
        mngr.add_key(xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem))
        mngr.load_cert_from_memory(self.load('rsacert.pem'), format=consts.KeyDataFormatPem, type=consts.KeyDataTypeTrusted)

    def test_load_cert_from_memory_with_bad_args(self):
        mngr = xmlsec.KeysManager()
        mngr.add_key(xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem))
        with self.assertRaises(TypeError):
            mngr.load_cert_from_memory(1, format='', type='')

    def test_load_cert_from_memory_invalid_data(self):
        mngr = xmlsec.KeysManager()
        mngr.add_key(xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem))
        with self.assertRaisesRegex(xmlsec.Error, '.*cannot load cert.*'):
            mngr.load_cert_from_memory(b'', format=consts.KeyDataFormatPem, type=consts.KeyDataTypeTrusted)

    def test_load_invalid_key(self):
        mngr = xmlsec.KeysManager()
        with self.assertRaises(ValueError):
            mngr.add_key(xmlsec.Key())



import contextlib
import json
from urllib.request import Request, urlopen

DEFAULT_USER_AGENT = 'https://github.com/xmlsec/python-xmlsec'
DOWNLOAD_USER_AGENT = 'python-xmlsec build'


def make_request(url, github_token=None, json_response=False):
    headers = {'User-Agent': DEFAULT_USER_AGENT}
    if github_token:
        headers['authorization'] = 'Bearer ' + github_token
    request = Request(url, headers=headers)
    with contextlib.closing(urlopen(request)) as response:
        charset = response.headers.get_content_charset() or 'utf-8'
        content = response.read().decode(charset)
        if json_response:
            return json.loads(content)
        return content


def download_lib(url, filename):
    request = Request(url, headers={'User-Agent': DOWNLOAD_USER_AGENT})
    with urlopen(request) as response, open(filename, 'wb') as target:
        while True:
            chunk = response.read(8192)
            if not chunk:
                break
            target.write(chunk)



import unittest

from lxml import etree

import xmlsec
from tests import base

consts = xmlsec.constants


class TestTemplates(base.TestMemoryLeaks):
    def test_create(self):
        root = self.load_xml('doc.xml')
        sign = xmlsec.template.create(
            root, c14n_method=consts.TransformExclC14N, sign_method=consts.TransformRsaSha1, id='Id', ns='test'
        )
        self.assertEqual('Id', sign.get('Id'))
        self.assertEqual('test', sign.prefix)

    def test_create_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.template.create('', c14n_method=consts.TransformExclC14N, sign_method=consts.TransformRsaSha1)

    def test_encrypt_data_create(self):
        root = self.load_xml('doc.xml')
        enc = xmlsec.template.encrypted_data_create(
            root, method=consts.TransformDes3Cbc, id='Id', type='Type', mime_type='MimeType', encoding='Encoding', ns='test'
        )
        for a in ('Id', 'Type', 'MimeType', 'Encoding'):
            self.assertEqual(a, enc.get(a))
        self.assertEqual('test', enc.prefix)

    def test_ensure_key_info(self):
        root = self.load_xml('doc.xml')
        sign = xmlsec.template.create(root, c14n_method=consts.TransformExclC14N, sign_method=consts.TransformRsaSha1)
        ki = xmlsec.template.ensure_key_info(sign, id='Id')
        self.assertEqual('Id', ki.get('Id'))

    def test_ensure_key_info_fail(self):
        with self.assertRaisesRegex(xmlsec.Error, 'cannot ensure key info.'):
            xmlsec.template.ensure_key_info(etree.fromstring(b'<Data/>'), id='Id')

    def test_ensure_key_info_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.template.ensure_key_info('', id=0)

    def test_add_encrypted_key(self):
        root = self.load_xml('doc.xml')
        sign = xmlsec.template.create(root, c14n_method=consts.TransformExclC14N, sign_method=consts.TransformRsaSha1)
        ki = xmlsec.template.ensure_key_info(sign)
        ek = xmlsec.template.add_encrypted_key(ki, consts.TransformRsaOaep)
        self.assertEqual(ek, xmlsec.tree.find_node(self.load_xml('sign_template.xml'), consts.NodeEncryptedKey, consts.EncNs))
        ek2 = xmlsec.template.add_encrypted_key(ki, consts.TransformRsaOaep, id='Id', type='Type', recipient='Recipient')
        for a in ('Id', 'Type', 'Recipient'):
            self.assertEqual(a, ek2.get(a))

    def test_add_key_name(self):
        root = self.load_xml('doc.xml')
        sign = xmlsec.template.create(root, c14n_method=consts.TransformExclC14N, sign_method=consts.TransformRsaSha1)
        ki = xmlsec.template.ensure_key_info(sign)
        kn = xmlsec.template.add_key_name(ki)
        self.assertEqual(kn, xmlsec.tree.find_node(self.load_xml('sign_template.xml'), consts.NodeKeyName, consts.DSigNs))
        kn2 = xmlsec.template.add_key_name(ki, name='name')
        self.assertEqual('name', kn2.text)

    def test_add_key_name_none(self):
        root = self.load_xml('doc.xml')
        sign = xmlsec.template.create(root, c14n_method=consts.TransformExclC14N, sign_method=consts.TransformRsaSha1)
        ki = xmlsec.template.ensure_key_info(sign)
        kn2 = xmlsec.template.add_key_name(ki, name=None)
        self.assertEqual(kn2.text, None)
        print(etree.tostring(kn2))

    def test_add_key_name_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.template.add_key_name('')

    def test_add_reference(self):
        root = self.load_xml('doc.xml')
        sign = xmlsec.template.create(root, c14n_method=consts.TransformExclC14N, sign_method=consts.TransformRsaSha1)
        ref = xmlsec.template.add_reference(sign, consts.TransformSha1, id='Id', uri='URI', type='Type')
        for a in ('Id', 'URI', 'Type'):
            self.assertEqual(a, ref.get(a))

    def test_add_reference_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.template.add_reference('', consts.TransformSha1)
        with self.assertRaises(TypeError):
            xmlsec.template.add_reference(etree.Element('root'), '')

    def test_add_reference_fail(self):
        with self.assertRaisesRegex(xmlsec.Error, 'cannot add reference.'):
            xmlsec.template.add_reference(etree.Element('root'), consts.TransformSha1)

    def test_add_transform_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.template.add_transform('', consts.TransformSha1)
        with self.assertRaises(TypeError):
            xmlsec.template.add_transform(etree.Element('root'), '')

    def test_add_key_value(self):
        root = self.load_xml('doc.xml')
        sign = xmlsec.template.create(root, c14n_method=consts.TransformExclC14N, sign_method=consts.TransformRsaSha1)
        ki = xmlsec.template.ensure_key_info(sign)
        kv = xmlsec.template.add_key_value(ki)
        self.assertEqual(kv, xmlsec.tree.find_node(self.load_xml('sign_template.xml'), consts.NodeKeyValue, consts.DSigNs))

    def test_add_key_value_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.template.add_key_value('')

    def test_add_x509_data(self):
        root = self.load_xml('doc.xml')
        sign = xmlsec.template.create(root, c14n_method=consts.TransformExclC14N, sign_method=consts.TransformRsaSha1)
        ki = xmlsec.template.ensure_key_info(sign)
        x509 = xmlsec.template.add_x509_data(ki)
        xmlsec.template.x509_data_add_certificate(x509)
        xmlsec.template.x509_data_add_crl(x509)
        issuer = xmlsec.template.x509_data_add_issuer_serial(x509)
        xmlsec.template.x509_data_add_ski(x509)
        xmlsec.template.x509_data_add_subject_name(x509)
        xmlsec.template.x509_issuer_serial_add_issuer_name(issuer)
        xmlsec.template.x509_issuer_serial_add_serial_number(issuer)
        self.assertEqual(x509, xmlsec.tree.find_node(self.load_xml('sign_template.xml'), consts.NodeX509Data, consts.DSigNs))

    def test_add_x509_data_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.template.add_x509_data('')

    def test_x509_issuer_serial_add_issuer(self):
        root = self.load_xml('doc.xml')
        sign = xmlsec.template.create(root, c14n_method=consts.TransformExclC14N, sign_method=consts.TransformRsaSha1)
        ki = xmlsec.template.ensure_key_info(sign)
        x509 = xmlsec.template.add_x509_data(ki)
        issuer = xmlsec.template.x509_data_add_issuer_serial(x509)
        name = xmlsec.template.x509_issuer_serial_add_issuer_name(issuer, name='Name')
        serial = xmlsec.template.x509_issuer_serial_add_serial_number(issuer, serial='Serial')
        self.assertEqual('Name', name.text)
        self.assertEqual('Serial', serial.text)

    def test_x509_issuer_serial_add_issuer_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.template.x509_data_add_issuer_serial('')

    def test_x509_issuer_serial_add_issuer_name_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.template.x509_issuer_serial_add_issuer_name('')

    def test_x509_issuer_serial_add_serial_number_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.template.x509_issuer_serial_add_serial_number('')

    def test_x509_data_add_subject_name_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.template.x509_data_add_subject_name('')

    def test_x509_data_add_ski_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.template.x509_data_add_ski('')

    def test_x509_data_add_certificate_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.template.x509_data_add_certificate('')

    def test_x509_data_add_crl_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.template.x509_data_add_crl('')

    def test_add_encrypted_key_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.template.add_encrypted_key('', 0)

    def test_encrypted_data_create_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.template.encrypted_data_create('', 0)

    def test_encrypted_data_ensure_cipher_value(self):
        root = self.load_xml('doc.xml')
        enc = xmlsec.template.encrypted_data_create(root, method=consts.TransformDes3Cbc)
        cv = xmlsec.template.encrypted_data_ensure_cipher_value(enc)
        self.assertEqual(cv, xmlsec.tree.find_node(self.load_xml('sign_template.xml'), consts.NodeCipherValue, consts.EncNs))

    def test_encrypted_data_ensure_cipher_value_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.template.encrypted_data_ensure_cipher_value('')

    def test_encrypted_data_ensure_key_info(self):
        root = self.load_xml('doc.xml')
        enc = xmlsec.template.encrypted_data_create(root, method=consts.TransformDes3Cbc)
        ki = xmlsec.template.encrypted_data_ensure_key_info(enc)
        self.assertEqual(ki, xmlsec.tree.find_node(self.load_xml('enc_template.xml'), consts.NodeKeyInfo, consts.DSigNs))
        ki2 = xmlsec.template.encrypted_data_ensure_key_info(enc, id='Id', ns='test')
        self.assertEqual('Id', ki2.get('Id'))
        self.assertEqual('test', ki2.prefix)

    def test_encrypted_data_ensure_key_info_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.template.encrypted_data_ensure_key_info('')

    @unittest.skipIf(not hasattr(consts, 'TransformXslt'), reason='XSLT transformations not enabled')
    def test_transform_add_c14n_inclusive_namespaces(self):
        root = self.load_xml('doc.xml')
        sign = xmlsec.template.create(root, c14n_method=consts.TransformExclC14N, sign_method=consts.TransformRsaSha1)
        ref = xmlsec.template.add_reference(sign, consts.TransformSha1)
        trans1 = xmlsec.template.add_transform(ref, consts.TransformEnveloped)
        xmlsec.template.transform_add_c14n_inclusive_namespaces(trans1, 'default')
        trans2 = xmlsec.template.add_transform(ref, consts.TransformXslt)
        xmlsec.template.transform_add_c14n_inclusive_namespaces(trans2, ['ns1', 'ns2'])
        self.assertEqual(ref, xmlsec.tree.find_node(self.load_xml('sign_template.xml'), consts.NodeReference, consts.DSigNs))

    def test_transform_add_c14n_inclusive_namespaces_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.template.transform_add_c14n_inclusive_namespaces('', [])



import xmlsec
from tests import base

consts = xmlsec.constants


class TestTree(base.TestMemoryLeaks):
    def test_find_child(self):
        root = self.load_xml('sign_template.xml')
        si = xmlsec.tree.find_child(root, consts.NodeSignedInfo, consts.DSigNs)
        self.assertEqual(consts.NodeSignedInfo, si.tag.partition('}')[2])
        self.assertIsNone(xmlsec.tree.find_child(root, consts.NodeReference))
        self.assertIsNone(xmlsec.tree.find_child(root, consts.NodeSignedInfo, consts.EncNs))

    def test_find_child_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.tree.find_child('', 0, True)

    def test_find_parent(self):
        root = self.load_xml('sign_template.xml')
        si = xmlsec.tree.find_child(root, consts.NodeSignedInfo, consts.DSigNs)
        self.assertIs(root, xmlsec.tree.find_parent(si, consts.NodeSignature))
        self.assertIsNone(xmlsec.tree.find_parent(root, consts.NodeSignedInfo))

    def test_find_parent_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.tree.find_parent('', 0, True)

    def test_find_node(self):
        root = self.load_xml('sign_template.xml')
        ref = xmlsec.tree.find_node(root, consts.NodeReference)
        self.assertEqual(consts.NodeReference, ref.tag.partition('}')[2])
        self.assertIsNone(xmlsec.tree.find_node(root, consts.NodeReference, consts.EncNs))

    def test_find_node_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.tree.find_node('', 0, True)

    def test_add_ids(self):
        root = self.load_xml('sign_template.xml')
        xmlsec.tree.add_ids(root, ['id1', 'id2', 'id3'])

    def test_add_ids_bad_args(self):
        with self.assertRaises(TypeError):
            xmlsec.tree.add_ids('', [])



"""Testing the PKCS#11 shim layer.

Heavily inspired by from https://github.com/IdentityPython/pyXMLSecurity by leifj
under license "As is", see https://github.com/IdentityPython/pyXMLSecurity/blob/master/LICENSE.txt
"""

import logging
import os
import shutil
import subprocess
import tempfile
import traceback
import unittest

DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')


def paths_for_component(component: str, default_paths):
    env_path = os.environ.get(component)
    return [env_path] if env_path else default_paths


def find_alts(component_name, alts) -> str:
    for a in alts:
        if os.path.exists(a):
            return a
    raise unittest.SkipTest(f'Required component is missing: {component_name}')


def run_cmd(args, softhsm_conf=None):
    env = {}
    if softhsm_conf is not None:
        env['SOFTHSM_CONF'] = softhsm_conf
        env['SOFTHSM2_CONF'] = softhsm_conf
    proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env)
    out, err = proc.communicate()
    if err is not None and len(err) > 0:
        logging.error(err)
    if out is not None and len(out) > 0:
        logging.debug(out)
    rv = proc.wait()
    if rv:
        with open(softhsm_conf) as f:
            conf = f.read()
        msg = '[cmd: {cmd}] [code: {code}] [stdout: {out}] [stderr: {err}] [config: {conf}]'
        msg = msg.format(
            cmd=' '.join(args),
            code=rv,
            out=out.strip(),
            err=err.strip(),
            conf=conf,
        )
        raise RuntimeError(msg)
    return out, err


component_default_paths = {
    'P11_MODULE': [
        '/usr/lib/softhsm/libsofthsm2.so',
        '/usr/lib/x86_64-linux-gnu/softhsm/libsofthsm2.so',
        '/usr/lib/softhsm/libsofthsm.so',
        '/usr/lib64/softhsm/libsofthsm2.so',
    ],
    'P11_ENGINE': [
        '/usr/lib/ssl/engines/libpkcs11.so',
        '/usr/lib/engines/engine_pkcs11.so',
        '/usr/lib/x86_64-linux-gnu/engines-1.1/pkcs11.so',
        '/usr/lib64/engines-1.1/pkcs11.so',
        '/usr/lib64/engines-1.1/libpkcs11.so',
        '/usr/lib64/engines-3/pkcs11.so',
        '/usr/lib64/engines-3/libpkcs11.so',
        '/usr/lib/x86_64-linux-gnu/engines-3/pkcs11.so',
        '/usr/lib/x86_64-linux-gnu/engines-3/libpkcs11.so',
    ],
    'PKCS11_TOOL': [
        '/usr/bin/pkcs11-tool',
    ],
    'SOFTHSM': [
        '/usr/bin/softhsm2-util',
        '/usr/bin/softhsm',
    ],
    'OPENSSL': [
        '/usr/bin/openssl',
    ],
}

component_path = {
    component_name: find_alts(component_name, paths_for_component(component_name, default_paths))
    for component_name, default_paths in component_default_paths.items()
}

softhsm_version = 1
if component_path['SOFTHSM'].endswith('softhsm2-util'):
    softhsm_version = 2

openssl_version = subprocess.check_output([component_path['OPENSSL'], 'version'])[8:11].decode()

p11_test_files = []
softhsm_conf = None
softhsm_db = None


def _temp_file() -> str:
    f = tempfile.NamedTemporaryFile(delete=False)  # noqa: SIM115
    p11_test_files.append(f.name)
    return f.name


def _temp_dir() -> str:
    d = tempfile.mkdtemp()
    p11_test_files.append(d)
    return d


@unittest.skipIf(component_path['P11_MODULE'] is None, 'SoftHSM PKCS11 module not installed')
def setup() -> None:
    logging.debug('Creating test pkcs11 token using softhsm')
    try:
        global softhsm_conf
        softhsm_conf = _temp_file()
        logging.debug('Generating softhsm.conf')
        with open(softhsm_conf, 'w') as f:
            if softhsm_version == 2:
                softhsm_db = _temp_dir()
                f.write(
                    f"""
# Generated by test
directories.tokendir = {softhsm_db}
objectstore.backend = file
log.level = DEBUG
"""
                )
            else:
                softhsm_db = _temp_file()
                f.write(
                    f"""
# Generated by test
0:{softhsm_db}
"""
                )

        logging.debug('Initializing the token')
        _, _ = run_cmd(
            [
                component_path['SOFTHSM'],
                '--slot',
                '0',
                '--label',
                'test',
                '--init-token',
                '--pin',
                'secret1',
                '--so-pin',
                'secret2',
            ],
            softhsm_conf=softhsm_conf,
        )

        hash_priv_key = _temp_file()
        logging.debug('Converting test private key to format for softhsm')
        run_cmd(
            [
                component_path['OPENSSL'],
                'pkcs8',
                '-topk8',
                '-inform',
                'PEM',
                '-outform',
                'PEM',
                '-nocrypt',
                '-in',
                os.path.join(DATA_DIR, 'rsakey.pem'),
                '-out',
                hash_priv_key,
            ],
            softhsm_conf=softhsm_conf,
        )

        logging.debug('Importing the test key to softhsm')
        run_cmd(
            [
                component_path['SOFTHSM'],
                '--import',
                hash_priv_key,
                '--token',
                'test',
                '--id',
                'a1b2',
                '--label',
                'test',
                '--pin',
                'secret1',
            ],
            softhsm_conf=softhsm_conf,
        )
        run_cmd(
            [
                component_path['PKCS11_TOOL'],
                '--module',
                component_path['P11_MODULE'],
                '-l',
                '--pin',
                'secret1',
                '-O',
            ],
            softhsm_conf=softhsm_conf,
        )
        signer_cert_pem = _temp_file()
        openssl_conf = _temp_file()
        logging.debug('Generating OpenSSL config for version %s', openssl_version)
        with open(openssl_conf, 'w') as f:
            f.write(
                '\n'.join(
                    [
                        'openssl_conf = openssl_def',
                        '[openssl_def]',
                        'engines = engine_section',
                        '[engine_section]',
                        'pkcs11 = pkcs11_section',
                        '[req]',
                        'distinguished_name = req_distinguished_name',
                        '[req_distinguished_name]',
                        '[pkcs11_section]',
                        'engine_id = pkcs11',
                        # dynamic_path,
                        'MODULE_PATH = {}'.format(component_path['P11_MODULE']),
                        'init = 0',
                    ]
                )
            )

        with open(openssl_conf) as f:
            logging.debug('-------- START DEBUG openssl_conf --------')
            logging.debug(f.readlines())
            logging.debug('-------- END DEBUG openssl_conf --------')
        logging.debug('-------- START DEBUG paths --------')
        logging.debug(run_cmd(['ls', '-ld', component_path['P11_ENGINE']]))
        logging.debug(run_cmd(['ls', '-ld', component_path['P11_MODULE']]))
        logging.debug('-------- END DEBUG paths --------')

        signer_cert_der = _temp_file()

        logging.debug('Generating self-signed certificate')
        run_cmd(
            [
                component_path['OPENSSL'],
                'req',
                '-new',
                '-x509',
                '-subj',
                '/CN=Test Signer',
                '-engine',
                'pkcs11',
                '-config',
                openssl_conf,
                '-keyform',
                'engine',
                '-key',
                'label_test',
                '-passin',
                'pass:secret1',
                '-out',
                signer_cert_pem,
            ],
            softhsm_conf=softhsm_conf,
        )

        run_cmd(
            [
                component_path['OPENSSL'],
                'x509',
                '-inform',
                'PEM',
                '-outform',
                'DER',
                '-in',
                signer_cert_pem,
                '-out',
                signer_cert_der,
            ],
            softhsm_conf=softhsm_conf,
        )

        logging.debug('Importing certificate into token')

        run_cmd(
            [
                component_path['PKCS11_TOOL'],
                '--module',
                component_path['P11_MODULE'],
                '-l',
                '--slot-index',
                '0',
                '--id',
                'a1b2',
                '--label',
                'test',
                '-y',
                'cert',
                '-w',
                signer_cert_der,
                '--pin',
                'secret1',
            ],
            softhsm_conf=softhsm_conf,
        )

        # TODO: Should be teardowned in teardown
        os.environ['SOFTHSM_CONF'] = softhsm_conf
        os.environ['SOFTHSM2_CONF'] = softhsm_conf

    except Exception as ex:
        print('-' * 64)
        traceback.print_exc()
        print('-' * 64)
        logging.exception('PKCS11 tests disabled: unable to initialize test token')
        raise ex


def teardown() -> None:
    global p11_test_files
    for o in p11_test_files:
        if os.path.exists(o):
            if os.path.isdir(o):
                shutil.rmtree(o)
            else:
                os.unlink(o)
    p11_test_files = []



import unittest

import xmlsec
from tests import base

consts = xmlsec.constants


class TestSignContext(base.TestMemoryLeaks):
    def test_init(self):
        ctx = xmlsec.SignatureContext(manager=xmlsec.KeysManager())
        del ctx

    def test_init_no_keys_manager(self):
        ctx = xmlsec.SignatureContext()
        del ctx

    def test_init_bad_args(self):
        with self.assertRaisesRegex(TypeError, 'KeysManager required'):
            xmlsec.SignatureContext(manager='foo')

    def test_no_key(self):
        ctx = xmlsec.SignatureContext(manager=xmlsec.KeysManager())
        self.assertIsNone(ctx.key)

    def test_del_key(self):
        ctx = xmlsec.SignatureContext(manager=xmlsec.KeysManager())
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        del ctx.key
        self.assertIsNone(ctx.key)

    def test_set_key(self):
        ctx = xmlsec.SignatureContext(manager=xmlsec.KeysManager())
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(ctx.key)

    def test_set_key_bad_type(self):
        ctx = xmlsec.SignatureContext(manager=xmlsec.KeysManager())
        with self.assertRaisesRegex(TypeError, r'instance of \*xmlsec.Key\* expected.'):
            ctx.key = ''

    def test_set_invalid_key(self):
        ctx = xmlsec.SignatureContext(manager=xmlsec.KeysManager())
        with self.assertRaisesRegex(TypeError, 'empty key.'):
            ctx.key = xmlsec.Key()

    def test_register_id(self):
        ctx = xmlsec.SignatureContext()
        root = self.load_xml('sign_template.xml')
        sign = xmlsec.template.create(root, consts.TransformExclC14N, consts.TransformRsaSha1, 'Id')
        ctx.register_id(sign, 'Id')

    def test_register_id_bad_args(self):
        ctx = xmlsec.SignatureContext()
        with self.assertRaises(TypeError):
            ctx.register_id('')

    def test_register_id_with_namespace_without_attribute(self):
        ctx = xmlsec.SignatureContext()
        root = self.load_xml('sign_template.xml')
        sign = xmlsec.template.create(root, consts.TransformExclC14N, consts.TransformRsaSha1, 'Id')
        with self.assertRaisesRegex(xmlsec.Error, 'missing attribute.'):
            ctx.register_id(sign, 'Id', id_ns='foo')

    def test_sign_bad_args(self):
        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        with self.assertRaises(TypeError):
            ctx.sign('')

    def test_sign_fail(self):
        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        with self.assertRaisesRegex(xmlsec.Error, 'failed to sign'):
            ctx.sign(self.load_xml('sign1-in.xml'))

    def test_sign_case1(self):
        """Should sign a pre-constructed template file using a key from a PEM file."""
        root = self.load_xml('sign1-in.xml')
        sign = xmlsec.tree.find_node(root, consts.NodeSignature)
        self.assertIsNotNone(sign)

        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(ctx.key)
        ctx.key.name = 'rsakey.pem'
        self.assertEqual('rsakey.pem', ctx.key.name)

        ctx.sign(sign)
        self.assertEqual(self.load_xml('sign1-out.xml'), root)

    def test_sign_case2(self):
        """Should sign a dynamicaly constructed template file using a key from a PEM file."""
        root = self.load_xml('sign2-in.xml')
        sign = xmlsec.template.create(root, consts.TransformExclC14N, consts.TransformRsaSha1)
        self.assertIsNotNone(sign)
        root.append(sign)
        ref = xmlsec.template.add_reference(sign, consts.TransformSha1)
        xmlsec.template.add_transform(ref, consts.TransformEnveloped)
        ki = xmlsec.template.ensure_key_info(sign)
        xmlsec.template.add_key_name(ki)

        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(ctx.key)
        ctx.key.name = 'rsakey.pem'
        self.assertEqual('rsakey.pem', ctx.key.name)

        ctx.sign(sign)
        self.assertEqual(self.load_xml('sign2-out.xml'), root)

    def test_sign_case3(self):
        """Should sign a file using a dynamicaly created template, key from PEM and an X509 cert."""
        root = self.load_xml('sign3-in.xml')
        sign = xmlsec.template.create(root, consts.TransformExclC14N, consts.TransformRsaSha1)
        self.assertIsNotNone(sign)
        root.append(sign)
        ref = xmlsec.template.add_reference(sign, consts.TransformSha1)
        xmlsec.template.add_transform(ref, consts.TransformEnveloped)
        ki = xmlsec.template.ensure_key_info(sign)
        xmlsec.template.add_x509_data(ki)

        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(ctx.key)
        ctx.key.load_cert_from_file(self.path('rsacert.pem'), consts.KeyDataFormatPem)
        ctx.key.name = 'rsakey.pem'
        self.assertEqual('rsakey.pem', ctx.key.name)

        ctx.sign(sign)
        self.assertEqual(self.load_xml('sign3-out.xml'), root)

    def test_sign_case4(self):
        """Should sign a file using a dynamically created template, key from PEM and an X509 cert with custom ns."""
        root = self.load_xml('sign4-in.xml')
        xmlsec.tree.add_ids(root, ['ID'])
        elem_id = root.get('ID', None)
        if elem_id:
            elem_id = '#' + elem_id
        sign = xmlsec.template.create(root, consts.TransformExclC14N, consts.TransformRsaSha1, ns='ds')
        self.assertIsNotNone(sign)
        root.append(sign)
        ref = xmlsec.template.add_reference(sign, consts.TransformSha1, uri=elem_id)
        xmlsec.template.add_transform(ref, consts.TransformEnveloped)
        xmlsec.template.add_transform(ref, consts.TransformExclC14N)
        ki = xmlsec.template.ensure_key_info(sign)
        xmlsec.template.add_x509_data(ki)

        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(ctx.key)
        ctx.key.load_cert_from_file(self.path('rsacert.pem'), consts.KeyDataFormatPem)
        ctx.key.name = 'rsakey.pem'
        self.assertEqual('rsakey.pem', ctx.key.name)

        ctx.sign(sign)
        self.assertEqual(self.load_xml('sign4-out.xml'), root)

    def test_sign_case5(self):
        """Should sign a file using a dynamicaly created template, key from PEM file and an X509 certificate."""
        root = self.load_xml('sign5-in.xml')
        sign = xmlsec.template.create(root, consts.TransformExclC14N, consts.TransformRsaSha1)
        self.assertIsNotNone(sign)
        root.append(sign)
        ref = xmlsec.template.add_reference(sign, consts.TransformSha1)
        xmlsec.template.add_transform(ref, consts.TransformEnveloped)

        ki = xmlsec.template.ensure_key_info(sign)
        x509 = xmlsec.template.add_x509_data(ki)
        xmlsec.template.x509_data_add_subject_name(x509)
        xmlsec.template.x509_data_add_certificate(x509)
        xmlsec.template.x509_data_add_ski(x509)
        x509_issuer_serial = xmlsec.template.x509_data_add_issuer_serial(x509)
        xmlsec.template.x509_issuer_serial_add_issuer_name(x509_issuer_serial, 'Test Issuer')
        xmlsec.template.x509_issuer_serial_add_serial_number(x509_issuer_serial, '1')

        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(ctx.key)
        ctx.key.load_cert_from_file(self.path('rsacert.pem'), consts.KeyDataFormatPem)
        ctx.key.name = 'rsakey.pem'
        self.assertEqual('rsakey.pem', ctx.key.name)

        ctx.sign(sign)
        if (1, 2, 36) <= xmlsec.get_libxmlsec_version() <= (1, 2, 37):
            expected_xml_file = 'sign5-out-xmlsec_1_2_36_to_37.xml'
        else:
            expected_xml_file = 'sign5-out.xml'
        self.assertEqual(self.load_xml(expected_xml_file), root)

    def test_sign_binary_bad_args(self):
        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        with self.assertRaises(TypeError):
            ctx.sign_binary(bytes=1, transform='')

    def test_sign_binary_no_key(self):
        ctx = xmlsec.SignatureContext()
        with self.assertRaisesRegex(xmlsec.Error, 'Sign key is not specified.'):
            ctx.sign_binary(bytes=b'', transform=consts.TransformRsaSha1)

    @unittest.skipIf(not hasattr(consts, 'TransformXslt'), reason='XSLT transformations not enabled')
    def test_sign_binary_invalid_signature_method(self):
        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        with self.assertRaisesRegex(xmlsec.Error, 'incompatible signature method'):
            ctx.sign_binary(bytes=b'', transform=consts.TransformXslt)

    def test_sign_binary(self):
        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(ctx.key)
        ctx.key.name = 'rsakey.pem'
        self.assertEqual('rsakey.pem', ctx.key.name)

        sign = ctx.sign_binary(self.load('sign6-in.bin'), consts.TransformRsaSha1)
        self.assertEqual(self.load('sign6-out.bin'), sign)

    def test_sign_binary_twice_not_possible(self):
        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        data = self.load('sign6-in.bin')
        ctx.sign_binary(data, consts.TransformRsaSha1)
        with self.assertRaisesRegex(xmlsec.Error, 'Signature context already used; it is designed for one use only.'):
            ctx.sign_binary(data, consts.TransformRsaSha1)

    def test_verify_bad_args(self):
        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        with self.assertRaises(TypeError):
            ctx.verify('')

    def test_verify_fail(self):
        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        with self.assertRaisesRegex(xmlsec.Error, 'failed to verify'):
            ctx.verify(self.load_xml('sign1-in.xml'))

    def test_verify_case_1(self):
        self.check_verify(1)

    def test_verify_case_2(self):
        self.check_verify(2)

    def test_verify_case_3(self):
        self.check_verify(3)

    def test_verify_case_4(self):
        self.check_verify(4)

    def test_verify_case_5(self):
        self.check_verify(5)

    def check_verify(self, i):
        root = self.load_xml(f'sign{i}-out.xml')
        xmlsec.tree.add_ids(root, ['ID'])
        sign = xmlsec.tree.find_node(root, consts.NodeSignature)
        self.assertIsNotNone(sign)
        self.assertEqual(consts.NodeSignature, sign.tag.partition('}')[2])

        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsapub.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(ctx.key)
        ctx.key.name = 'rsapub.pem'
        self.assertEqual('rsapub.pem', ctx.key.name)
        ctx.verify(sign)

    def test_validate_binary_sign(self):
        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(ctx.key)
        ctx.key.name = 'rsakey.pem'
        self.assertEqual('rsakey.pem', ctx.key.name)

        ctx.verify_binary(self.load('sign6-in.bin'), consts.TransformRsaSha1, self.load('sign6-out.bin'))

    def test_validate_binary_sign_fail(self):
        ctx = xmlsec.SignatureContext()

        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        self.assertIsNotNone(ctx.key)
        ctx.key.name = 'rsakey.pem'
        self.assertEqual('rsakey.pem', ctx.key.name)
        with self.assertRaises(xmlsec.Error):
            ctx.verify_binary(self.load('sign6-in.bin'), consts.TransformRsaSha1, b'invalid')

    def test_enable_reference_transform(self):
        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        ctx.enable_reference_transform(consts.TransformRsaSha1)

    def test_enable_reference_transform_bad_args(self):
        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        with self.assertRaises(TypeError):
            ctx.enable_reference_transform('')
        with self.assertRaises(TypeError):
            ctx.enable_reference_transform(0)
        with self.assertRaises(TypeError):
            ctx.enable_reference_transform(consts.KeyDataAes)

    def test_enable_signature_transform(self):
        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        ctx.enable_signature_transform(consts.TransformRsaSha1)

    def test_enable_signature_transform_bad_args(self):
        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        with self.assertRaises(TypeError):
            ctx.enable_signature_transform('')
        with self.assertRaises(TypeError):
            ctx.enable_signature_transform(0)
        with self.assertRaises(TypeError):
            ctx.enable_signature_transform(consts.KeyDataAes)

    def test_set_enabled_key_data(self):
        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        ctx.set_enabled_key_data([consts.KeyDataAes])

    def test_set_enabled_key_data_empty(self):
        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        ctx.set_enabled_key_data([])

    def test_set_enabled_key_data_bad_args(self):
        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        with self.assertRaises(TypeError):
            ctx.set_enabled_key_data(0)

    def test_set_enabled_key_data_bad_list(self):
        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        with self.assertRaisesRegex(TypeError, 'expected list of KeyData constants.'):
            ctx.set_enabled_key_data('foo')



import sys
from io import BytesIO
from unittest import skipIf

import xmlsec
from tests import base
from xmlsec import constants as consts


class TestBase64LineSize(base.TestMemoryLeaks):
    def tearDown(self):
        xmlsec.base64_default_line_size(64)
        super().tearDown()

    def test_get_base64_default_line_size(self):
        self.assertEqual(xmlsec.base64_default_line_size(), 64)

    def test_set_base64_default_line_size_positional_arg(self):
        xmlsec.base64_default_line_size(0)
        self.assertEqual(xmlsec.base64_default_line_size(), 0)

    def test_set_base64_default_line_size_keyword_arg(self):
        xmlsec.base64_default_line_size(size=0)
        self.assertEqual(xmlsec.base64_default_line_size(), 0)

    def test_set_base64_default_line_size_with_bad_args(self):
        size = xmlsec.base64_default_line_size()
        for bad_size in (None, '', object()):
            with self.assertRaises(TypeError):
                xmlsec.base64_default_line_size(bad_size)
        self.assertEqual(xmlsec.base64_default_line_size(), size)

    def test_set_base64_default_line_size_rejects_negative_values(self):
        size = xmlsec.base64_default_line_size()
        with self.assertRaises(ValueError):
            xmlsec.base64_default_line_size(-1)
        self.assertEqual(xmlsec.base64_default_line_size(), size)


class TestCallbacks(base.TestMemoryLeaks):
    def setUp(self):
        super().setUp()
        xmlsec.cleanup_callbacks()

    def _sign_doc(self):
        root = self.load_xml('doc.xml')
        sign = xmlsec.template.create(root, c14n_method=consts.TransformExclC14N, sign_method=consts.TransformRsaSha1)
        xmlsec.template.add_reference(sign, consts.TransformSha1, uri='cid:123456')

        ctx = xmlsec.SignatureContext()
        ctx.key = xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem)
        ctx.sign(sign)
        return sign

    def _expect_sign_failure(self):
        with self.assertRaisesRegex(xmlsec.Error, 'failed to sign'):
            self._sign_doc()

    def _mismatch_callbacks(self, match_cb=lambda filename: False):
        return [
            match_cb,
            lambda filename: None,
            lambda none, buf: 0,
            lambda none: None,
        ]

    def _register_mismatch_callbacks(self, match_cb=lambda filename: False):
        xmlsec.register_callbacks(*self._mismatch_callbacks(match_cb))

    def _register_match_callbacks(self):
        xmlsec.register_callbacks(
            lambda filename: filename == b'cid:123456',
            lambda filename: BytesIO(b'<html><head/><body/></html>'),
            lambda bio, buf: bio.readinto(buf),
            lambda bio: bio.close(),
        )

    def _find(self, elem, *tags):
        try:
            return elem.xpath(
                './' + '/'.join(f'xmldsig:{tag}' for tag in tags),
                namespaces={
                    'xmldsig': 'http://www.w3.org/2000/09/xmldsig#',
                },
            )[0]
        except IndexError as e:
            raise KeyError(tags) from e

    def _verify_external_data_signature(self):
        signature = self._sign_doc()
        digest = self._find(signature, 'SignedInfo', 'Reference', 'DigestValue').text
        self.assertEqual(digest, 'VihZwVMGJ48NsNl7ertVHiURXk8=')

    def test_sign_external_data_no_callbacks_fails(self):
        self._expect_sign_failure()

    def test_sign_external_data_default_callbacks_fails(self):
        xmlsec.register_default_callbacks()
        self._expect_sign_failure()

    def test_sign_external_data_no_matching_callbacks_fails(self):
        self._register_mismatch_callbacks()
        self._expect_sign_failure()

    def test_sign_data_from_callbacks(self):
        self._register_match_callbacks()
        self._verify_external_data_signature()

    def test_sign_data_not_first_callback(self):
        bad_match_calls = 0

        def match_cb(filename):
            nonlocal bad_match_calls
            bad_match_calls += 1
            return False

        for _ in range(2):
            self._register_mismatch_callbacks(match_cb)

        self._register_match_callbacks()

        for _ in range(2):
            self._register_mismatch_callbacks()

        self._verify_external_data_signature()
        self.assertEqual(bad_match_calls, 0)

    @skipIf(sys.platform == 'win32', 'unclear behaviour on windows')
    def test_failed_sign_because_default_callbacks(self):
        mismatch_calls = 0

        def mismatch_cb(filename):
            nonlocal mismatch_calls
            mismatch_calls += 1
            return False

        # NB: These first two sets of callbacks should never get called,
        # because the default callbacks always match beforehand:
        self._register_match_callbacks()
        self._register_mismatch_callbacks(mismatch_cb)
        xmlsec.register_default_callbacks()
        self._register_mismatch_callbacks(mismatch_cb)
        self._register_mismatch_callbacks(mismatch_cb)
        self._expect_sign_failure()
        self.assertEqual(mismatch_calls, 2)

    def test_register_non_callables(self):
        for idx in range(4):
            cbs = self._mismatch_callbacks()
            cbs[idx] = None
            self.assertRaises(TypeError, xmlsec.register_callbacks, *cbs)

    def test_sign_external_data_fails_on_read_callback_wrong_returns(self):
        xmlsec.register_callbacks(
            lambda filename: filename == b'cid:123456',
            lambda filename: BytesIO(b'<html><head/><body/></html>'),
            lambda bio, buf: None,
            lambda bio: bio.close(),
        )
        self._expect_sign_failure()



import gc
import os
import sys
import unittest

from lxml import etree

import xmlsec

etype = type(etree.Element('test'))

ns = {'dsig': xmlsec.constants.DSigNs, 'enc': xmlsec.constants.EncNs}


try:
    import resource

    test_iterations = int(os.environ.get('PYXMLSEC_TEST_ITERATIONS', '10'))
except (ImportError, ValueError):
    test_iterations = 0


class TestMemoryLeaks(unittest.TestCase):
    maxDiff = None

    iterations = test_iterations

    data_dir = os.path.join(os.path.dirname(__file__), 'data')

    def setUp(self):
        gc.disable()
        self.addTypeEqualityFunc(etype, 'assertXmlEqual')
        xmlsec.enable_debug_trace(1)

    def run(self, result=None):
        # run first time
        super().run(result=result)
        if self.iterations == 0:
            return

        m_usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        o_count = gc.get_count()[0]
        m_hits = 0
        o_hits = 0
        for _ in range(self.iterations):
            super().run(result=result)
            m_usage_n = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
            if m_usage_n > m_usage:
                m_usage = m_usage_n
                m_hits += 1
            o_count_n = gc.get_count()[0]
            if o_count_n > o_count:
                o_count = o_count_n
                o_hits += 1
            del m_usage_n
            del o_count_n

        if m_hits > int(self.iterations * 0.8):
            result.buffer = False
            try:
                raise AssertionError('memory leak detected')
            except AssertionError:
                result.addError(self, sys.exc_info())
        if o_hits > int(self.iterations * 0.8):
            result.buffer = False
            try:
                raise AssertionError('unreferenced objects detected')
            except AssertionError:
                result.addError(self, sys.exc_info())

    def path(self, name):
        """Return full path for resource."""
        return os.path.join(self.data_dir, name)

    def load(self, name):
        """Load resource by name."""
        with open(self.path(name), 'rb') as stream:
            return stream.read()

    def load_xml(self, name, xpath=None):
        """Return xml.etree."""
        with open(self.path(name)) as f:
            root = etree.parse(f).getroot()
            if xpath is None:
                return root
            return root.find(xpath)

    def dump(self, root):
        print(etree.tostring(root))

    def assertXmlEqual(self, first, second, msg=None):
        """Check equality of etree.roots."""
        msg = msg or ''
        if first.tag != second.tag:
            self.fail(f'Tags do not match: {first.tag} and {second.tag}. {msg}')
        for name, value in first.attrib.items():
            if second.attrib.get(name) != value:
                self.fail(f'Attributes do not match: {name}={value!r}, {name}={second.attrib.get(name)!r}. {msg}')
        for name in second.attrib:
            if name not in first.attrib:
                self.fail(f'x2 has an attribute x1 is missing: {name}. {msg}')
        if not _xml_text_compare(first.text, second.text):
            self.fail(f'text: {first.text!r} != {second.text!r}. {msg}')
        if not _xml_text_compare(first.tail, second.tail):
            self.fail(f'tail: {first.tail!r} != {second.tail!r}. {msg}')
        cl1 = sorted(first.getchildren(), key=lambda x: x.tag)
        cl2 = sorted(second.getchildren(), key=lambda x: x.tag)
        if len(cl1) != len(cl2):
            self.fail(f'children length differs, {len(cl1)} != {len(cl2)}. {msg}')
        for c1, c2 in zip(cl1, cl2):
            self.assertXmlEqual(c1, c2)


def _xml_text_compare(t1, t2):
    if not t1 and not t2:
        return True
    if t1 == '*' or t2 == '*':
        return True
    return (t1 or '').strip() == (t2 or '').strip()



import tempfile

from lxml import etree

import xmlsec
from tests import base

consts = xmlsec.constants


class TestEncryptionContext(base.TestMemoryLeaks):
    def test_init(self):
        ctx = xmlsec.EncryptionContext(manager=xmlsec.KeysManager())
        del ctx

    def test_init_no_keys_manager(self):
        ctx = xmlsec.EncryptionContext()
        del ctx

    def test_init_bad_args(self):
        with self.assertRaisesRegex(TypeError, 'KeysManager required'):
            xmlsec.EncryptionContext(manager='foo')

    def test_no_key(self):
        ctx = xmlsec.EncryptionContext(manager=xmlsec.KeysManager())
        self.assertIsNone(ctx.key)

    def test_get_key(self):
        ctx = xmlsec.EncryptionContext(manager=xmlsec.KeysManager())
        self.assertIsNone(ctx.key)
        ctx.key = xmlsec.Key.from_file(self.path('rsacert.pem'), format=consts.KeyDataFormatCertPem)
        self.assertIsNotNone(ctx.key)

    def test_del_key(self):
        ctx = xmlsec.EncryptionContext(manager=xmlsec.KeysManager())
        ctx.key = xmlsec.Key.from_file(self.path('rsacert.pem'), format=consts.KeyDataFormatCertPem)
        del ctx.key
        self.assertIsNone(ctx.key)

    def test_set_key(self):
        ctx = xmlsec.EncryptionContext(manager=xmlsec.KeysManager())
        ctx.key = xmlsec.Key.from_file(self.path('rsacert.pem'), format=consts.KeyDataFormatCertPem)
        self.assertIsNotNone(ctx.key)

    def test_set_key_bad_type(self):
        ctx = xmlsec.EncryptionContext(manager=xmlsec.KeysManager())
        with self.assertRaisesRegex(TypeError, r'instance of \*xmlsec.Key\* expected.'):
            ctx.key = ''

    def test_set_invalid_key(self):
        ctx = xmlsec.EncryptionContext(manager=xmlsec.KeysManager())
        with self.assertRaisesRegex(TypeError, 'empty key.'):
            ctx.key = xmlsec.Key()

    def test_encrypt_xml(self):
        root = self.load_xml('enc1-in.xml')
        enc_data = xmlsec.template.encrypted_data_create(root, consts.TransformAes128Cbc, type=consts.TypeEncElement, ns='xenc')
        xmlsec.template.encrypted_data_ensure_cipher_value(enc_data)
        ki = xmlsec.template.encrypted_data_ensure_key_info(enc_data, ns='dsig')
        ek = xmlsec.template.add_encrypted_key(ki, consts.TransformRsaOaep)
        xmlsec.template.encrypted_data_ensure_cipher_value(ek)
        data = root.find('./Data')
        self.assertIsNotNone(data)

        manager = xmlsec.KeysManager()
        manager.add_key(xmlsec.Key.from_file(self.path('rsacert.pem'), format=consts.KeyDataFormatCertPem))

        ctx = xmlsec.EncryptionContext(manager)
        ctx.key = xmlsec.Key.generate(consts.KeyDataAes, 128, consts.KeyDataTypeSession)

        encrypted = ctx.encrypt_xml(enc_data, data)
        self.assertIsNotNone(encrypted)

        enc_method = xmlsec.tree.find_child(enc_data, consts.NodeEncryptionMethod, consts.EncNs)
        self.assertIsNotNone(enc_method)
        self.assertEqual('http://www.w3.org/2001/04/xmlenc#aes128-cbc', enc_method.get('Algorithm'))
        ki = xmlsec.tree.find_child(enc_data, consts.NodeKeyInfo, consts.DSigNs)
        self.assertIsNotNone(ki)
        enc_method2 = xmlsec.tree.find_node(ki, consts.NodeEncryptionMethod, consts.EncNs)
        self.assertIsNotNone(enc_method2)
        self.assertEqual('http://www.w3.org/2001/04/xmlenc#rsa-oaep-mgf1p', enc_method2.get('Algorithm'))
        cipher_value = xmlsec.tree.find_node(ki, consts.NodeCipherValue, consts.EncNs)
        self.assertIsNotNone(cipher_value)

    def test_encrypt_xml_bad_args(self):
        ctx = xmlsec.EncryptionContext()
        with self.assertRaises(TypeError):
            ctx.encrypt_xml('', 0)

    def test_encrypt_xml_bad_template(self):
        ctx = xmlsec.EncryptionContext()
        with self.assertRaisesRegex(xmlsec.Error, 'unsupported `Type`, it should be `element` or `content`'):
            ctx.encrypt_xml(etree.Element('root'), etree.Element('node'))

    def test_encrypt_xml_bad_template_bad_type_attribute(self):
        ctx = xmlsec.EncryptionContext()
        with self.assertRaisesRegex(xmlsec.Error, 'unsupported `Type`, it should be `element` or `content`'):
            root = etree.Element('root')
            root.attrib['Type'] = 'foo'
            ctx.encrypt_xml(root, etree.Element('node'))

    def test_encrypt_xml_fail(self):
        ctx = xmlsec.EncryptionContext()
        with self.assertRaisesRegex(xmlsec.Error, 'failed to encrypt xml'):
            root = etree.Element('root')
            root.attrib['Type'] = consts.TypeEncElement
            ctx.encrypt_xml(root, etree.Element('node'))

    def test_encrypt_binary(self):
        root = self.load_xml('enc2-in.xml')
        enc_data = xmlsec.template.encrypted_data_create(
            root, consts.TransformAes128Cbc, type=consts.TypeEncContent, ns='xenc', mime_type='binary/octet-stream'
        )
        xmlsec.template.encrypted_data_ensure_cipher_value(enc_data)
        ki = xmlsec.template.encrypted_data_ensure_key_info(enc_data, ns='dsig')
        ek = xmlsec.template.add_encrypted_key(ki, consts.TransformRsaOaep)
        xmlsec.template.encrypted_data_ensure_cipher_value(ek)

        manager = xmlsec.KeysManager()
        manager.add_key(xmlsec.Key.from_file(self.path('rsacert.pem'), format=consts.KeyDataFormatCertPem))

        ctx = xmlsec.EncryptionContext(manager)
        ctx.key = xmlsec.Key.generate(consts.KeyDataAes, 128, consts.KeyDataTypeSession)

        encrypted = ctx.encrypt_binary(enc_data, b'test')
        self.assertIsNotNone(encrypted)
        self.assertEqual(f'{{{consts.EncNs}}}{consts.NodeEncryptedData}', encrypted.tag)

        enc_method = xmlsec.tree.find_child(enc_data, consts.NodeEncryptionMethod, consts.EncNs)
        self.assertIsNotNone(enc_method)
        self.assertEqual('http://www.w3.org/2001/04/xmlenc#aes128-cbc', enc_method.get('Algorithm'))

        ki = xmlsec.tree.find_child(enc_data, consts.NodeKeyInfo, consts.DSigNs)
        self.assertIsNotNone(ki)
        enc_method2 = xmlsec.tree.find_node(ki, consts.NodeEncryptionMethod, consts.EncNs)
        self.assertIsNotNone(enc_method2)
        self.assertEqual('http://www.w3.org/2001/04/xmlenc#rsa-oaep-mgf1p', enc_method2.get('Algorithm'))
        cipher_value = xmlsec.tree.find_node(ki, consts.NodeCipherValue, consts.EncNs)
        self.assertIsNotNone(cipher_value)

    def test_encrypt_binary_bad_args(self):
        ctx = xmlsec.EncryptionContext()
        with self.assertRaises(TypeError):
            ctx.encrypt_binary('', 0)

    def test_encrypt_binary_bad_template(self):
        ctx = xmlsec.EncryptionContext()
        with self.assertRaisesRegex(xmlsec.Error, 'failed to encrypt binary'):
            ctx.encrypt_binary(etree.Element('root'), b'data')

    def test_encrypt_uri(self):
        root = self.load_xml('enc2-in.xml')
        enc_data = xmlsec.template.encrypted_data_create(
            root, consts.TransformAes128Cbc, type=consts.TypeEncContent, ns='xenc', mime_type='binary/octet-stream'
        )
        xmlsec.template.encrypted_data_ensure_cipher_value(enc_data)
        ki = xmlsec.template.encrypted_data_ensure_key_info(enc_data, ns='dsig')
        ek = xmlsec.template.add_encrypted_key(ki, consts.TransformRsaOaep)
        xmlsec.template.encrypted_data_ensure_cipher_value(ek)

        manager = xmlsec.KeysManager()
        manager.add_key(xmlsec.Key.from_file(self.path('rsacert.pem'), format=consts.KeyDataFormatCertPem))

        ctx = xmlsec.EncryptionContext(manager)
        ctx.key = xmlsec.Key.generate(consts.KeyDataAes, 128, consts.KeyDataTypeSession)

        with tempfile.NamedTemporaryFile(delete=False) as tmpfile:
            tmpfile.write(b'test')

        encrypted = ctx.encrypt_binary(enc_data, 'file://' + tmpfile.name)
        self.assertIsNotNone(encrypted)
        self.assertEqual(f'{{{consts.EncNs}}}{consts.NodeEncryptedData}', encrypted.tag)

        enc_method = xmlsec.tree.find_child(enc_data, consts.NodeEncryptionMethod, consts.EncNs)
        self.assertIsNotNone(enc_method)
        self.assertEqual('http://www.w3.org/2001/04/xmlenc#aes128-cbc', enc_method.get('Algorithm'))

        ki = xmlsec.tree.find_child(enc_data, consts.NodeKeyInfo, consts.DSigNs)
        self.assertIsNotNone(ki)
        enc_method2 = xmlsec.tree.find_node(ki, consts.NodeEncryptionMethod, consts.EncNs)
        self.assertIsNotNone(enc_method2)
        self.assertEqual('http://www.w3.org/2001/04/xmlenc#rsa-oaep-mgf1p', enc_method2.get('Algorithm'))
        cipher_value = xmlsec.tree.find_node(ki, consts.NodeCipherValue, consts.EncNs)
        self.assertIsNotNone(cipher_value)

    def test_encrypt_uri_bad_args(self):
        ctx = xmlsec.EncryptionContext()
        with self.assertRaises(TypeError):
            ctx.encrypt_uri('', 0)

    def test_encrypt_uri_fail(self):
        ctx = xmlsec.EncryptionContext()
        with self.assertRaisesRegex(xmlsec.Error, 'failed to encrypt URI'):
            ctx.encrypt_uri(etree.Element('root'), '')

    def test_decrypt1(self):
        self.check_decrypt(1)

    def test_decrypt2(self):
        self.check_decrypt(2)

    def test_decrypt_key(self):
        root = self.load_xml('enc3-out.xml')
        enc_key = xmlsec.tree.find_child(root, consts.NodeEncryptedKey, consts.EncNs)
        self.assertIsNotNone(enc_key)

        manager = xmlsec.KeysManager()
        manager.add_key(xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem))
        ctx = xmlsec.EncryptionContext(manager)
        keydata = ctx.decrypt(enc_key)
        ctx.reset()
        root.remove(enc_key)
        ctx.key = xmlsec.Key.from_binary_data(consts.KeyDataAes, keydata)
        enc_data = xmlsec.tree.find_child(root, consts.NodeEncryptedData, consts.EncNs)
        self.assertIsNotNone(enc_data)
        decrypted = ctx.decrypt(enc_data)
        self.assertIsNotNone(decrypted)
        self.assertEqual(self.load_xml('enc3-in.xml'), decrypted)

    def check_decrypt(self, i):
        root = self.load_xml(f'enc{i}-out.xml')
        enc_data = xmlsec.tree.find_child(root, consts.NodeEncryptedData, consts.EncNs)
        self.assertIsNotNone(enc_data)

        manager = xmlsec.KeysManager()
        manager.add_key(xmlsec.Key.from_file(self.path('rsakey.pem'), format=consts.KeyDataFormatPem))
        ctx = xmlsec.EncryptionContext(manager)
        decrypted = ctx.decrypt(enc_data)
        self.assertIsNotNone(decrypted)
        self.assertEqual(self.load_xml(f'enc{i}-in.xml'), root)

    def test_decrypt_bad_args(self):
        ctx = xmlsec.EncryptionContext()
        with self.assertRaises(TypeError):
            ctx.decrypt('')



import os
import sys
from distutils import log
from distutils.errors import DistutilsError

from setuptools.command.build_ext import build_ext as build_ext_orig

from .static_build import CrossCompileInfo, StaticBuildHelper


class build_ext(build_ext_orig):
    def info(self, message):
        self.announce(message, level=log.INFO)

    def run(self):
        ext = self.ext_map['xmlsec']
        self.debug = os.environ.get('PYXMLSEC_ENABLE_DEBUG', False)
        self.static = os.environ.get('PYXMLSEC_STATIC_DEPS', False)
        self.size_opt = os.environ.get('PYXMLSEC_OPTIMIZE_SIZE', True)

        if self.static or sys.platform == 'win32':
            helper = StaticBuildHelper(self)
            helper.prepare(sys.platform)
        else:
            import pkgconfig

            try:
                config = pkgconfig.parse('xmlsec1')
            except OSError as error:
                raise DistutilsError('Unable to invoke pkg-config.') from error
            except pkgconfig.PackageNotFoundError as error:
                raise DistutilsError('xmlsec1 is not installed or not in path.') from error

            if config is None or not config.get('libraries'):
                raise DistutilsError('Bad or incomplete result returned from pkg-config.')

            ext.define_macros.extend(config['define_macros'])
            ext.include_dirs.extend(config['include_dirs'])
            ext.library_dirs.extend(config['library_dirs'])
            ext.libraries.extend(config['libraries'])

        import lxml

        ext.include_dirs.extend(lxml.get_include())

        ext.define_macros.extend(
            [('MODULE_NAME', self.distribution.metadata.name), ('MODULE_VERSION', self.distribution.metadata.version)]
        )
        for key, value in ext.define_macros:
            if key == 'XMLSEC_CRYPTO' and not (value.startswith('"') and value.endswith('"')):
                ext.define_macros.remove((key, value))
                ext.define_macros.append((key, f'"{value}"'))
                break

        if sys.platform == 'win32':
            ext.extra_compile_args.append('/Zi')
        else:
            ext.extra_compile_args.extend(
                [
                    '-g',
                    '-std=c99',
                    '-fPIC',
                    '-fno-strict-aliasing',
                    '-Wno-error=declaration-after-statement',
                    '-Werror=implicit-function-declaration',
                ]
            )

        if self.debug:
            ext.define_macros.append(('PYXMLSEC_ENABLE_DEBUG', '1'))
            if sys.platform == 'win32':
                ext.extra_compile_args.append('/Od')
            else:
                ext.extra_compile_args.append('-Wall')
                ext.extra_compile_args.append('-O0')
        else:
            if self.size_opt:
                if sys.platform == 'win32':
                    ext.extra_compile_args.append('/Os')
                else:
                    ext.extra_compile_args.append('-Os')

        super().run()


__all__ = ('CrossCompileInfo', 'build_ext')



"""Test constants from :mod:`xmlsec.constants` module."""

import pytest

import xmlsec


def _constants(typename):
    return list(
        sorted(
            (
                getattr(xmlsec.constants, name)
                for name in dir(xmlsec.constants)
                if type(getattr(xmlsec.constants, name)).__name__ == typename
            ),
            key=lambda t: t.name.lower(),
        )
    )


@pytest.mark.parametrize('transform', _constants('__Transform'), ids=repr)
def test_transform_str(transform):
    """Test string representation of ``xmlsec.constants.__Transform``."""
    assert str(transform) == f'{transform.name}, {transform.href}'


@pytest.mark.parametrize('transform', _constants('__Transform'), ids=repr)
def test_transform_repr(transform):
    """Test raw string representation of ``xmlsec.constants.__Transform``."""
    assert repr(transform) == f'__Transform({transform.name!r}, {transform.href!r}, {transform.usage})'


@pytest.mark.parametrize('keydata', _constants('__KeyData'), ids=repr)
def test_keydata_str(keydata):
    """Test string representation of ``xmlsec.constants.__KeyData``."""
    assert str(keydata) == f'{keydata.name}, {keydata.href}'


@pytest.mark.parametrize('keydata', _constants('__KeyData'), ids=repr)
def test_keydata_repr(keydata):
    """Test raw string representation of ``xmlsec.constants.__KeyData``."""
    assert repr(keydata) == f'__KeyData({keydata.name!r}, {keydata.href!r})'



import html.parser
import os
import re
from distutils import log
from distutils.version import StrictVersion as Version

from .network import make_request


class HrefCollector(html.parser.HTMLParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hrefs = []

    def handle_starttag(self, tag, attrs):
        if tag == 'a':
            for name, value in attrs:
                if name == 'href':
                    self.hrefs.append(value)


def latest_release_from_html(url, matcher):
    content = make_request(url)
    collector = HrefCollector()
    collector.feed(content)
    hrefs = collector.hrefs

    def comp(text):
        try:
            return Version(matcher.match(text).groupdict()['version'])
        except (AttributeError, ValueError):
            return Version('0.0')

    latest = max(hrefs, key=comp)
    return f'{url}/{latest}'


def latest_release_from_gnome_org_cache(url, lib_name):
    cache_url = f'{url}/cache.json'
    cache = make_request(cache_url, json_response=True)
    latest_version = cache[2][lib_name][-1]
    latest_source = cache[1][lib_name][latest_version]['tar.xz']
    return f'{url}/{latest_source}'


def latest_release_json_from_github_api(repo):
    api_url = f'https://api.github.com/repos/{repo}/releases/latest'
    token = os.environ.get('GH_TOKEN')
    if token:
        log.info('Using GitHub token to avoid rate limiting')
    return make_request(api_url, token, json_response=True)


def latest_openssl_release():
    return latest_release_json_from_github_api('openssl/openssl')['tarball_url']


def latest_zlib_release():
    return latest_release_from_html('https://zlib.net/fossils', re.compile('zlib-(?P<version>.*).tar.gz'))


def latest_libiconv_release():
    return latest_release_from_html('https://ftpmirror.gnu.org/libiconv', re.compile('libiconv-(?P<version>.*).tar.gz'))


def latest_libxml2_release():
    return latest_release_from_gnome_org_cache('https://download.gnome.org/sources/libxml2', 'libxml2')


def latest_libxslt_release():
    return latest_release_from_gnome_org_cache('https://download.gnome.org/sources/libxslt', 'libxslt')


def latest_xmlsec_release():
    assets = latest_release_json_from_github_api('lsh123/xmlsec')['assets']
    (tar_gz,) = [asset for asset in assets if asset['name'].endswith('.tar.gz')]
    return tar_gz['browser_download_url']



from pathlib import Path

from setuptools import Extension, setup

from build_support.build_ext import build_ext

src_root = Path(__file__).parent / 'src'
sources = [str(path.absolute()) for path in src_root.rglob('*.c')]
pyxmlsec = Extension('xmlsec', sources=sources)
setup_reqs = ['setuptools_scm[toml]>=3.4', 'pkgconfig>=1.5.1', 'lxml>=3.8']


with open('README.md', encoding='utf-8') as readme:
    long_desc = readme.read()


setup(
    name='xmlsec',
    use_scm_version=True,
    description='Python bindings for the XML Security Library',
    long_description=long_desc,
    long_description_content_type='text/markdown',
    ext_modules=[pyxmlsec],
    cmdclass={'build_ext': build_ext},
    python_requires='>=3.9',
    setup_requires=setup_reqs,
    install_requires=['lxml>=3.8'],
    author='Bulat Gaifullin',
    author_email='support@mehcode.com',
    maintainer='Oleg Hoefling',
    maintainer_email='oleg.hoefling@gmail.com',
    url='https://github.com/mehcode/python-xmlsec',
    project_urls={
        'Documentation': 'https://xmlsec.readthedocs.io',
        'Source': 'https://github.com/mehcode/python-xmlsec',
        'Changelog': 'https://github.com/mehcode/python-xmlsec/releases',
    },
    license='MIT',
    keywords=['xmlsec'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: C',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
        'Topic :: Text Processing :: Markup :: XML',
        'Typing :: Typed',
    ],
    zip_safe=False,
    packages=['xmlsec'],
    package_dir={'': 'src'},
    package_data={'xmlsec': ['py.typed', '*.pyi']},
)



import xmlsec
from tests import base


class TestModule(base.TestMemoryLeaks):
    def test_reinitialize_module(self):
        """This test doesn't explicitly verify anything, but will be invoked first in the suite.

        So if the subsequent tests don't fail, we know that the ``init()``/``shutdown()``
        function pair doesn't break anything.
        """
        xmlsec.shutdown()
        xmlsec.init()
